"""
Utility classes and functions to handle connection to a libvirt host system

The entire contents of callables in this module (minus the names defined in
NOCLOSE below), will become methods of the Virtadmin and VirtadminPersistent classes.
A Closure class is used to wrap the module functions, lambda does not
properly store instance state in this implementation.

Because none of the methods have a 'self' parameter defined, the classes
are defined to be dict-like, and get passed in to the methods as a the
special ``**dargs`` parameter.  All virtadmin module functions _MUST_ include a
special ``**dargs`` (variable keyword arguments) to accept non-default
keyword arguments.

The standard set of keyword arguments to all functions/modules is declared
in the VirtadminBase class.  Only the 'virtadmin_exec' key is guaranteed to always
be present, the remainder may or may not be provided.  Therefor, virtadmin
functions/methods should use the dict.get() method to retrieve with a default
for non-existant keys.

:copyright: 2012 Red Hat Inc.
"""

import signal
import logging
import re
import weakref
import time
import select

import aexpect
from aexpect import remote

from avocado.utils import path
from avocado.utils import process

from . import propcan
from . import utils_misc


# list of symbol names NOT to wrap as Virtadmin class methods
# Everything else from globals() will become a method of Virtadmin class
NOCLOSE = list(globals().keys()) + [
    'NOCLOSE', 'SCREENSHOT_ERROR_COUNT', 'VIRTADMIN_COMMAND_CACHE',
    'VIRTADMIN_EXEC', 'VirtadminBase', 'VirtadminClosure', 'VirtadminSession', 'Virtadmin',
    'VirtadminPersistent', 'VirtadminConnectBack', 'VIRTADMIN_COMMAND_GROUP_CACHE',
    'VIRTADMIN_COMMAND_GROUP_CACHE_NO_DETAIL',
]

# Needs to be in-scope for Virtadmin* class screenshot method and module function
SCREENSHOT_ERROR_COUNT = 0

# Cache of virtadmin commands, used by help_command_group() and help_command_only()
# TODO: Make the cache into a class attribute on VirtadminBase class.
VIRTADMIN_COMMAND_CACHE = None
VIRTADMIN_COMMAND_GROUP_CACHE = None
VIRTADMIN_COMMAND_GROUP_CACHE_NO_DETAIL = False

# This is used both inside and outside classes
try:
    VIRTADMIN_EXEC = path.find_command("virt-admin")
except path.CmdNotFoundError:
    logging.warning("virt-admin executable not set or found on path, "
                    "virtadmin-admin module will not function normally")
    VIRTADMIN_EXEC = '/bin/true'


class VirtadminBase(propcan.PropCanBase):

    """
    Base Class storing libvirt Connection & state to a host
    """

    __slots__ = ('uri', 'ignore_status', 'debug', 'virtadmin_exec', 'readonly')

    def __init__(self, *args, **dargs):
        """
        Initialize instance with virtadmin_exec always set to something
        """
        init_dict = dict(*args, **dargs)
        init_dict['virtadmin_exec'] = init_dict.get('virtadmin_exec', VIRTADMIN_EXEC)
        init_dict['uri'] = init_dict.get('uri', None)
        init_dict['debug'] = init_dict.get('debug', False)
        init_dict['ignore_status'] = init_dict.get('ignore_status', False)
        init_dict['readonly'] = init_dict.get('readonly', False)
        super(VirtadminBase, self).__init__(init_dict)

    def get_uri(self):
        """
        Accessor method for 'uri' property that must exist
        """
        # self.get() would call get_uri() recursivly
        try:
            return self.__dict_get__('uri')
        except KeyError:
            return None


class VirtadminSession(aexpect.ShellSession):

    """
    A virtadmin shell session, used with Virtadmin instances.
    """

    # No way to get virtadmin sub-command "exit" status
    # Check output against list of known error-status strings
    ERROR_REGEX_LIST = ['error:\s*.+$', '.*failed.*']

    def __init__(self, virtadmin_exec=None, uri=None, a_id=None,
                 prompt=r"virt-admin\s*[\#\>]\s*", remote_ip=None,
                 remote_user=None, remote_pwd=None,
                 ssh_remote_auth=False, readonly=False,
                 unprivileged_user=None,
                 auto_close=False, check_libvirtd=True):
        """
        Initialize virtadmin session server, or client if id set.

        :param virtadmin_exec: path to virtadmin executable
        :param uri: uri of libvirt instance to connect to
        :param id: ID of an already running server, if accessing a running
                server, or None if starting a new one.
        :param prompt: Regular expression describing the shell's prompt line.
        :param remote_ip: Hostname/IP of remote system to ssh into (if any)
        :param remote_user: Username to ssh in as (if any)
        :param remote_pwd: Password to use, or None for host/pubkey
        :param auto_close: Param to init ShellSession.
        :param ssh_remote_auth: ssh to remote first.(VirtadminConnectBack).
                                Then execute virtadmin commands.

        Because the VirtadminSession is designed for class VirtadminPersistent, so
        the default value of auto_close is False, and we manage the reference
        to VirtadminSession in VirtadminPersistent manually with counter_increase and
        counter_decrease. If you really want to use it directly over VirtadminPe-
        rsistent, please init it with auto_close=True, then the session will
        be closed in __del__.

            * session = VirtadminSession(virtadmin.VIRSH_EXEC, auto_close=True)
        """

        self.uri = uri
        self.remote_ip = remote_ip
        self.remote_user = remote_user
        self.remote_pwd = remote_pwd

        # Special handling if setting up a remote session
        if ssh_remote_auth:  # remote to remote
            logging.error("remote session is not supported by virt-admin yet.")
            if remote_pwd:
                pref_auth = "-o PreferredAuthentications=password"
            else:
                pref_auth = "-o PreferredAuthentications=hostbased,publickey"
            # ssh_cmd is not None flags this as remote session
            ssh_cmd = ("ssh -o UserKnownHostsFile=/dev/null %s -p %s %s@%s"
                       % (pref_auth, 22, self.remote_user, self.remote_ip))
            if uri:
                self.virtadmin_exec = ("%s \"%s -c '%s'\""
                                       % (ssh_cmd, virtadmin_exec, self.uri))
            else:
                self.virtadmin_exec = ("%s \"%s\"" % (ssh_cmd, virtadmin_exec))
        else:  # setting up a local session or re-using a session
            self.virtadmin_exec = virtadmin_exec
            if self.uri:
                self.virtadmin_exec += " -c '%s'" % self.uri
            ssh_cmd = None  # flags not-remote session

        if unprivileged_user:
            self.virtadmin_exec = "su - %s -c '%s'" % (unprivileged_user,
                                                       self.virtadmin_exec)

        # aexpect tries to auto close session because no clients connected yet
        aexpect.ShellSession.__init__(self, self.virtadmin_exec, a_id,
                                      prompt=prompt, auto_close=auto_close)

        # Handle remote session prompts:
        # 1.remote to remote with ssh
        # 2.local to remote with "virtadmin -c uri"
        if ssh_remote_auth or self.uri:
            # Handle ssh / password prompts
            remote.handle_prompts(self, self.remote_user, self.remote_pwd,
                                  prompt, debug=True)

        # fail if libvirtd is not running
        if check_libvirtd:
            if self.cmd_status('uri', timeout=60) != 0:
                logging.debug("Persistent virt-admin session is not responding, "
                              "libvirtd may be dead.")
                self.auto_close = True
                raise aexpect.ShellStatusError(virtadmin_exec, 'uri')

    def cmd_status_output(self, cmd, timeout=60, internal_timeout=None,
                          print_func=None, safe=False):
        """
        Send a virtadmin command and return its exit status and output.

        :param cmd: virtadmin command to send (must not contain newline characters)
        :param timeout: The duration (in seconds) to wait for the prompt to
                return
        :param internal_timeout: The timeout to pass to read_nonblocking
        :param print_func: A function to be used to print the data being read
                (should take a string parameter)
        :param safe: Whether using safe mode when execute cmd.
                In serial sessions, frequently the kernel might print debug or
                error messages that make read_up_to_prompt to timeout. Let's
                try to be a little more robust and send a carriage return, to
                see if we can get to the prompt when safe=True.

        :return: A tuple (status, output) where status is the exit status and
                output is the output of cmd
        :raise ShellTimeoutError: Raised if timeout expires
        :raise ShellProcessTerminatedError: Raised if the shell process
                terminates while waiting for output
        :raise ShellStatusError: Raised if the exit status cannot be obtained
        :raise ShellError: Raised if an unknown error occurs
        """
        out = self.cmd_output(cmd, timeout, internal_timeout, print_func, safe)
        for line in out.splitlines():
            if self.match_patterns(line, self.ERROR_REGEX_LIST) is not None:
                return 1, out
        return 0, out

    def cmd_result(self, cmd, ignore_status=False, debug=False, timeout=60):
        """
        Mimic process.run()
        :param cmd: virtadmin command to send.
        :param timeout: Time we'll wait until the process is finished.
        :returns: The command result object.
        """
        exit_status, stdout = self.cmd_status_output(cmd, timeout=timeout)
        stderr = ''  # no way to retrieve this separately
        result = process.CmdResult(cmd, stdout, stderr, exit_status)
        result.stdout = result.stdout_text
        result.stderr = result.stderr_text
        if not ignore_status and exit_status:
            raise process.CmdError(cmd, result,
                                   "Virtadmin Command returned non-zero exit status")
        if debug:
            logging.debug(result)
        return result

    def read_until_output_matches(self, patterns, filter_func=lambda x: x,
                                  timeout=60, internal_timeout=None,
                                  print_func=None, match_func=None):
        """
        Read from child using read_nonblocking until a pattern matches.

        Read using read_nonblocking until a match is found using match_patterns,
        or until timeout expires. Before attempting to search for a match, the
        data is filtered using the filter_func function provided.

        :param patterns: List of strings (regular expression patterns)
        :param filter_func: Function to apply to the data read from the child before
                attempting to match it against the patterns (should take and
                return a string)
        :param timeout: The duration (in seconds) to wait until a match is
                found
        :param internal_timeout: The timeout to pass to read_nonblocking
        :param print_func: A function to be used to print the data being read
                (should take a string parameter)
        :param match_func: Function to compare the output and patterns.
        :return: Tuple containing the match index and the data read so far
        :raise ExpectTimeoutError: Raised if timeout expires
        :raise ExpectProcessTerminatedError: Raised if the child process
                terminates while waiting for output
        :raise ExpectError: Raised if an unknown error occurs
        """
        if not match_func:
            match_func = self.match_patterns
        fd = self._get_fd("expect")
        o = ""
        end_time = time.time() + timeout
        while True:
            try:
                r, w, x = select.select([fd], [], [],
                                        max(0, end_time - time.time()))
            except (select.error, TypeError):
                break
            if not r:
                raise aexpect.ExpectTimeoutError(patterns, o)
            # Read data from child
            data = self.read_nonblocking(internal_timeout,
                                         end_time - time.time())
            if not data:
                break
            # Print it if necessary
            if print_func:
                for line in data.splitlines():
                    print_func(line)
            # Look for patterns
            o += data

            out = ''
            match = match_func(filter_func(o), patterns)
            if match is not None:
                output = o.splitlines()
                # Find the second match in output reverse list, only return
                # the content between the last match and the second last match.
                # read_nonblocking might include output of last command or help
                # info when session initiated,
                # e.g.
                # When use VirtadminPersistent initiate a virtadmin session, an list
                # command is send in to test libvirtd status, and the first
                # command output will be like:
                # Welcome to virtadmin, the virtualization interactive terminal.
                #
                # Type:  'help' for help with commands
                #       'quit' to quit
                #
                # virtadmin #  Id    Name                           State
                #----------------------------------------------------
                #
                # virtadmin #
                # the session help info is included, and the exact output
                # should be the content start after first virtadmin # prompt.
                # The list command did no harm here with help info included,
                # but sometime other commands get list command output included,
                # e.g.
                #  Running virtadmin command: net-list --all
                #  Sending command: net-list --all
                #  Id    Name                           State
                #  ----------------------------------------------------
                #
                # virtadmin #  Name            State      Autostart     Persistent
                #  ----------------------------------------------------------
                #  default              active     yes           yes
                #
                # virtadmin #
                # The list command output is mixed in the net-list command
                # output, this will fail to extract network name if use set
                # number 2 in list of output splitlines like in function
                # virtadmin.net_state_dict.
                for i in reversed(list(range(len(output) - 1))):
                    if match_func(output[i].strip(), patterns) is not None:
                        if re.split(patterns[match], output[i])[-1]:
                            output[i] = re.split(patterns[match],
                                                 output[i])[-1]
                            output_slice = output[i:]
                        else:
                            output_slice = output[i + 1:]
                        for j in range(len(output_slice) - 1):
                            output_slice[j] = output_slice[j] + '\n'
                        for k in range(len(output_slice)):
                            out += output_slice[k]
                        return match, out
                return match, o

        # Check if the child has terminated
        if utils_misc.wait_for(lambda: not self.is_alive(), 5, 0, 0.1):
            raise aexpect.ExpectProcessTerminatedError(patterns,
                                                       self.get_status(), o)
        else:
            # This shouldn't happen
            raise aexpect.ExpectError(patterns, o)


# Work around for inconsistent builtin closure local reference problem
# across different versions of python
class VirtadminClosure(object):

    """
    Callable with weak ref. to override ``**dargs`` when calling reference_function
    """

    def __init__(self, reference_function, dict_like_instance):
        """
        Callable reference_function with weak ref dict_like_instance
        """
        if not issubclass(dict_like_instance.__class__, dict):
            raise ValueError("dict_like_instance %s must be dict or subclass"
                             % dict_like_instance.__class__.__name__)
        self.reference_function = reference_function
        self.dict_like_weakref = weakref.ref(dict_like_instance)

    def __call__(self, *args, **dargs):
        """
        Call reference_function with dict_like_instance augmented by **dargs

        :param args: Passthrough to reference_function
        :param dargs: Updates dict_like_instance copy before call
        """
        new_dargs = self.dict_like_weakref()
        if new_dargs is None:
            new_dargs = {}
        for key in list(new_dargs.keys()):
            if key not in list(dargs.keys()):
                dargs[key] = new_dargs[key]
        return self.reference_function(*args, **dargs)


class Virtadmin(VirtadminBase):

    """
    Execute libvirt operations, using a new virtadmin shell each time.
    """

    __slots__ = []

    def __init__(self, *args, **dargs):
        """
        Initialize Virtadmin instance with persistent options

        :param args: Initial property keys/values
        :param dargs: Initial property keys/values
        """
        super(Virtadmin, self).__init__(*args, **dargs)
        # Define the instance callables from the contents of this module
        # to avoid using class methods and hand-written aliases
        for sym, ref in list(globals().items()):
            if sym not in NOCLOSE and callable(ref):
                # Adding methods, not properties, so avoid special __slots__
                # handling.  __getattribute__ will still find these.
                self.__super_set__(sym, VirtadminClosure(ref, self))


class VirtadminPersistent(Virtadmin):

    """
    Execute libvirt operations using persistent virtadmin session.
    """

    __slots__ = ('session_id', 'remote_pwd', 'remote_user', 'uri',
                 'remote_ip', 'ssh_remote_auth', 'unprivileged_user',
                 'readonly')

    # B/c the auto_close of VirtadminSession is False, we
    # need to manage the ref-count of it manually.
    COUNTERS = {}

    def __init__(self, *args, **dargs):
        super(VirtadminPersistent, self).__init__(*args, **dargs)
        if self.get('session_id') is None:
            # set_uri does not call when INITIALIZED = False
            # and no session_id passed to super __init__
            self.new_session()
        # increase the counter of session_id in COUNTERS.
        self.counter_increase()

    def __del__(self):
        """
        Clean up any leftover sessions
        """
        self.close_session()

    def counter_increase(self):
        """
        Method to increase the counter to self.a_id in COUNTERS.
        """
        session_id = self.__dict_get__("session_id")
        try:
            counter = self.__class__.COUNTERS[session_id]
        except KeyError as e:
            VirtadminPersistent.COUNTERS[session_id] = 1
            return
        # increase the counter of session_id.
        VirtadminPersistent.COUNTERS[session_id] += 1

    def counter_decrease(self):
        """
        Method to decrease the counter to self.a_id in COUNTERS.
        If the counter is less than 1, it means there is no more
        VirtadminSession instance referring to the session. So close
        this session, and return True.
        Else, decrease the counter in COUNTERS and return False.
        """
        session_id = self.__dict_get__("session_id")
        self.__class__.COUNTERS[session_id] -= 1
        counter = self.__class__.COUNTERS[session_id]
        if counter <= 0:
            # The last reference to this session. Closing it.
            session = VirtadminSession(a_id=session_id)
            # try nicely first
            session.close()
            if session.is_alive():
                # Be mean, in case it's hung
                session.close(sig=signal.SIGTERM)
            del self.__class__.COUNTERS[session_id]
            return True
        else:
            return False

    def close_session(self):
        """
        If a persistent session exists, close it down.
        """
        try:
            session_id = self.__dict_get__('session_id')
            if session_id:
                try:
                    existing = VirtadminSession(a_id=session_id)
                    if existing.is_alive():
                        self.counter_decrease()
                except (aexpect.ShellStatusError,
                        aexpect.ShellProcessTerminatedError):
                    # session was already closed
                    pass  # don't check is_alive or update counter
                self.__dict_del__("session_id")
        except KeyError:
            # Allow other exceptions to be raised
            pass  # session was closed already

    def new_session(self):
        """
        Open new session, closing any existing
        """
        # Accessors may call this method, avoid recursion
        # Must exist, can't be None
        virtadmin_exec = self.__dict_get__('virtadmin_exec')
        uri = self.__dict_get__('uri')  # Must exist, can be None
        readonly = self.__dict_get__('readonly')
        try:
            remote_user = self.__dict_get__('remote_user')
        except KeyError:
            remote_user = "root"
        try:
            remote_pwd = self.__dict_get__('remote_pwd')
        except KeyError:
            remote_pwd = None
        try:
            remote_ip = self.__dict_get__('remote_ip')
        except KeyError:
            remote_ip = None
        try:
            ssh_remote_auth = self.__dict_get__('ssh_remote_auth')
        except KeyError:
            ssh_remote_auth = False
        try:
            unprivileged_user = self.__dict_get__('unprivileged_user')
        except KeyError:
            unprivileged_user = None

        self.close_session()
        # Always create new session
        new_session = VirtadminSession(virtadmin_exec, uri, a_id=None,
                                       remote_ip=remote_ip,
                                       remote_user=remote_user,
                                       remote_pwd=remote_pwd,
                                       ssh_remote_auth=ssh_remote_auth,
                                       unprivileged_user=unprivileged_user,
                                       readonly=readonly)
        session_id = new_session.get_id()
        self.__dict_set__('session_id', session_id)

    def set_uri(self, uri):
        """
        Accessor method for 'uri' property, create new session on change
        """
        if not self.INITIALIZED:
            # Allow __init__ to call new_session
            self.__dict_set__('uri', uri)
        else:
            # If the uri is changing
            if self.__dict_get__('uri') != uri:
                self.__dict_set__('uri', uri)
                self.new_session()
            # otherwise do nothing


class VirtadminConnectBack(VirtadminPersistent):

    """
    Persistent virtadmin session connected back from a remote host
    """

    __slots__ = ('remote_ip', )

    def new_session(self):
        """
        Open new remote session, closing any existing
        """

        # Accessors may call this method, avoid recursion
        # Must exist, can't be None
        virtadmin_exec = self.__dict_get__('virtadmin_exec')
        uri = self.__dict_get__('uri')  # Must exist, can be None
        remote_ip = self.__dict_get__('remote_ip')
        try:
            remote_user = self.__dict_get__('remote_user')
        except KeyError:
            remote_user = 'root'
        try:
            remote_pwd = self.__dict_get__('remote_pwd')
        except KeyError:
            remote_pwd = None
        super(VirtadminConnectBack, self).close_session()
        new_session = VirtadminSession(virtadmin_exec, uri, a_id=None,
                                       remote_ip=remote_ip,
                                       remote_user=remote_user,
                                       remote_pwd=remote_pwd,
                                       ssh_remote_auth=True)
        session_id = new_session.get_id()
        self.__dict_set__('session_id', session_id)

    @staticmethod
    def kosher_args(remote_ip, uri):
        """
        Convenience static method to help validate argument sanity before use

        :param remote_ip: ip/hostname of remote libvirt helper-system
        :param uri: fully qualified libvirt uri of local system, from remote.
        :return: True/False if checks pass or not
        """
        if remote_ip is None or uri is None:
            return False
        all_false = [
            # remote_ip checks
            bool(remote_ip.count("EXAMPLE.COM")),
            bool(remote_ip.count("localhost")),
            bool(remote_ip.count("127.")),
            # uri checks
            uri is None,
            uri == "",
            bool(uri.count("default")),
            bool(uri.count(':///')),
            bool(uri.count("localhost")),
            bool(uri.count("127."))
        ]
        return True not in all_false


# virtadmin module functions follow (See module docstring for API) #####


def command(cmd, **dargs):
    """
    Interface to cmd function as 'cmd' symbol is polluted.

    :param cmd: Command line to append to virtadmin command
    :param dargs: standardized virtadmin function API keywords
    :return: CmdResult object
    :raise: CmdError if non-zero exit status and ignore_status=False
    """

    virtadmin_exec = dargs.get('virtadmin_exec', VIRTADMIN_EXEC)
    uri = dargs.get('uri', None)
    debug = dargs.get('debug', False)
    # Caller deals with errors
    ignore_status = dargs.get('ignore_status', True)
    session_id = dargs.get('session_id', None)
    readonly = dargs.get('readonly', False)
    unprivileged_user = dargs.get('unprivileged_user', None)
    timeout = dargs.get('timeout', None)

    # Check if this is a VirtadminPersistent method call
    if session_id:
        # Retrieve existing session
        session = VirtadminSession(a_id=session_id)
    else:
        session = None

    if debug:
        logging.debug("Running virtadmin command: %s", cmd)

    if timeout:
        try:
            timeout = int(timeout)
        except ValueError:
            logging.error("Ignore the invalid timeout value: %s", timeout)
            timeout = None

    if session:
        # Utilize persistent virtadmin session, not suit for readonly mode
        if readonly:
            logging.debug("Ignore readonly flag for this virtadmin session")
        if timeout is None:
            timeout = 60
        ret = session.cmd_result(cmd, ignore_status=ignore_status,
                                 debug=debug, timeout=timeout)
        # Mark return value with session it came from
        ret.from_session_id = session_id
    else:
        # Normal call to run virtadmin command
        # Readonly mode
        if readonly:
            logging.error("readonly mode is not supported by virt-admin yet.")
#            cmd = " -r " + cmd

        if uri:
            # uri argument IS being used
            uri_arg = " -c '%s' " % uri
        else:
            uri_arg = " "  # No uri argument being used

        cmd = "%s%s%s" % (virtadmin_exec, uri_arg, cmd)

        if unprivileged_user:
            # Run cmd as unprivileged user
            cmd = "su - %s -c '%s'" % (unprivileged_user, cmd)

        # Raise exception if ignore_status is False
        ret = process.run(cmd, timeout=timeout, verbose=debug,
                          ignore_status=ignore_status,
                          shell=True)
        # Mark return as not coming from persistent virtadmin session
        ret.from_session_id = None

    # Always log debug info, if persistent session or not
    if debug:
        logging.debug("status: %s", ret.exit_status)
        logging.debug("stdout: %s", ret.stdout_text.strip())
        logging.debug("stderr: %s", ret.stderr_text.strip())

    # Return CmdResult instance when ignore_status is True
    return ret


def cd(dir_path, options="", **dargs):
    """
    Run cd command in virsh interactive session.

    :param dir_path: dir path string
    :param options: extra options
    :param dargs: standardized virsh function API keywords
    :return: CmdResult object
    """
    cmd = "cd --dir %s %s" % (dir_path, options)
    return command(cmd, **dargs)


def pwd(options="", **dargs):
    """
    Run pwd command in virsh session.

    :param options: extra options
    :param dargs: standardized virsh function API keywords
    :return: CmdResult object
    """
    cmd = "pwd %s" % options
    return command(cmd, **dargs)


def echo(echo_str, options="", **dargs):
    """
    Run echo command in virsh session.

    :param echo_str: the echo string
    :param options: extra options
    :param dargs: standardized virsh function API keywords
    :return: CmdResult object
    """
    cmd = "echo %s %s" % (echo_str, options)
    return command(cmd, **dargs)


def exit(**dargs):
    """
    Run exit command in virsh session.

    :param dargs: standardized virsh function API keywords
    :return: CmdResult object
    """
    cmd = "exit"
    return command(cmd, **dargs)


def quit(**dargs):
    """
    Run quit command in virsh session.

    :param dargs: standardized virsh function API keywords
    :return: CmdResult object
    """
    cmd = "quit"
    return command(cmd, **dargs)


def srv_list(**dargs):
    """
    Run srv_list command in virt_admin session.

    :param dargs: standardized virt-admin function API keywords
    :return: CmdResult object
    """
    cmd = "srv-list"
    return command(cmd, **dargs)


def srv_threadpool_info(server_name, **dargs):
    """
    Run srv-threadpool-info in virt-admin session.
    :param server_name: list the attributes of the server with this name.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "srv-threadpool-info %s" % server_name
    return command(cmd, **dargs)


def srv_clients_info(server_name, **dargs):
    """
    Run srv-clients-info in virt-admin session
    :param server_name: list the attributes of the server with this name.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "srv-clients-info %s " % server_name
    return command(cmd, **dargs)


def srv_clients_list(server_name, **dargs):
    """
    Run srv-clients-list in virt-admin session
    :param server_name: list clients connected to this server.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "srv-clients-list %s " % server_name
    return command(cmd, **dargs)


def client_info(server_name, client_id, **dargs):
    """
    Run client-info in virt-admin session
    :param server_name: list clients connected to this server.
    :param client_id: client id number.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "client-info %s %s " % (server_name, client_id)
    return command(cmd, **dargs)


def srv_threadpool_set(server_name, min_workers=None,
                       max_workers=None, prio_workers=None,
                       options=None, **dargs):
    """
    Run srv-threadpool-set in virt-admin session
    :param server_name: set the workpool parameters of this server.
    :param min_workers: set the bottom limit to the number of workers.
    :param max_workers: set the upper limit to the number of workers.
    :param prio_workers: change the current number of priority workers.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "srv-threadpool-set %s" % server_name
    if min_workers:
        cmd += " --min-workers %s" % min_workers
    if max_workers:
        cmd += " --max-workers %s" % max_workers
    if prio_workers:
        cmd += " --priority-workers %s" % prio_workers
    if options:
        cmd += " %s" % options
    return command(cmd, **dargs)


def srv_clients_set(server_name, max_unauth_clients=None,
                    max_clients=None, options=None, **dargs):
    """
    Run srv-clients-set: set server's client-related configuration limits.
    :param max_unauth_clients: set the upper limit to number of clients
    for authentication to be connected to the server
    :param max_clients: set the upper limit to overall number of clients
    connected to the server.
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "srv-clients-set %s" % server_name
    if max_unauth_clients:
        cmd += " --max-unauth-clients %s" % max_unauth_clients
    if max_clients:
        cmd += " --max-clients %s" % max_clients
    if options:
        cmd += " %s" % options
    return command(cmd, **dargs)


def client_disconnect(server_name, client_id, **dargs):
    """
    Run client-disconnect: force disconnect a client from the given server
    :param server_name: the name of the server the client is currently connected to
    :param client_id: close a connection originating from client client_id
    :param dargs: standardized virt-admin function API keywords.
    :return: CmdResult object
    """
    cmd = "client-disconnect %s %s" % (server_name, client_id)
    return command(cmd, **dargs)



"""
graphics framebuffer device support class(es)

http://libvirt.org/formatdomain.html#elementsGraphics
"""

from virttest.libvirt_xml import accessors, vm_xml
from virttest.libvirt_xml.devices import base


class Graphics(base.TypedDeviceBase):

    __slots__ = ('passwd', 'channel', 'listen', 'listens', 'autoport', 'port',
                 'tlsPort', 'defaultMode', 'image_compression',
                 'jpeg_compression', 'zlib_compression', 'playback_compression',
                 'listen_type', 'listen_addr', 'passwdValidTo', 'clipboard_copypaste',
                 'filetransfer_enable', 'streaming_mode')

    def __init__(self, type_name='vnc', virsh_instance=base.base.virsh):
        # Add additional attribute 'passwd' for security
        accessors.XMLAttribute('passwd', self, parent_xpath='/',
                               tag_name='graphics', attribute='passwd')
        accessors.XMLAttribute('passwdValidTo', self, parent_xpath='/',
                               tag_name='graphics', attribute='passwdValidTo')
        accessors.XMLAttribute('listen', self, parent_xpath='/',
                               tag_name='graphics', attribute='listen')
        accessors.XMLAttribute('autoport', self, parent_xpath='/',
                               tag_name='graphics', attribute='autoport')
        accessors.XMLAttribute('port', self, parent_xpath='/',
                               tag_name='graphics', attribute='port')
        accessors.XMLAttribute('tlsPort', self, parent_xpath='/',
                               tag_name='graphics', attribute='tlsPort')
        accessors.XMLAttribute('type', self, parent_xpath='/',
                               tag_name='graphics', attribute='type')
        accessors.XMLAttribute('defaultMode', self, parent_xpath='/',
                               tag_name='graphics', attribute='defaultMode')
        accessors.XMLAttribute('image_compression', self, parent_xpath='/',
                               tag_name='image', attribute='compression')
        accessors.XMLAttribute('jpeg_compression', self, parent_xpath='/',
                               tag_name='jpeg', attribute='compression')
        accessors.XMLAttribute('zlib_compression', self, parent_xpath='/',
                               tag_name='zlib', attribute='compression')
        accessors.XMLAttribute('playback_compression', self, parent_xpath='/',
                               tag_name='playback', attribute='compression')
        accessors.XMLAttribute('listen_type', self, parent_xpath='/',
                               tag_name='listen', attribute='type')
        accessors.XMLAttribute('listen_addr', self, parent_xpath='/',
                               tag_name='listen', attribute='address')
        accessors.XMLAttribute('clipboard_copypaste', self, parent_xpath='/',
                               tag_name='clipboard', attribute='copypaste')
        accessors.XMLAttribute('filetransfer_enable', self, parent_xpath='/',
                               tag_name='filetransfer', attribute='enable')
        accessors.XMLAttribute('streaming_mode', self, parent_xpath='/',
                               tag_name='streaming', attribute='mode')
        super(Graphics, self).__init__(device_tag='graphics',
                                       type_name=type_name,
                                       virsh_instance=virsh_instance)

    def get_channel(self):
        """
        Return a list of dictionaries containing each channel's attributes
        """
        return self._get_list('channel')

    def set_channel(self, value):
        """
        Set all channel to the value list of dictionaries of channel attributes
        """
        self._set_list('channel', value)

    def del_channel(self):
        """
        Remove the list of dictionaries containing each channel's attributes
        """
        self._del_list('channel')

    def add_channel(self, **attributes):
        """
        Convenience method for appending channel from dictionary of attributes
        """
        self._add_item('channel', **attributes)

    def get_listens(self):
        """
        Return a list of dictionaries containing each listen's attributes
        """
        return self._get_list('listen')

    def set_listens(self, value):
        """
        Set all listens to the value list of dictionaries of listen attributes
        """
        self._set_list('listen', value)

    def del_listens(self):
        """
        Remove the list of dictionaries containing each listen's attributes
        """
        self._del_list('listen')

    def add_listens(self, **attributes):
        """
        Convenience method for appending listens from dictionary of attributes
        """
        self._add_item('listen', **attributes)

    @staticmethod
    def change_graphic_type_passwd(vm_name, graphic, passwd=None):
        """
        Change the graphic type name and passwd

        :param vm_name: name of vm
        :param graphic: graphic type, spice or vnc
        :param passwd: password for graphic
        """
        vmxml = vm_xml.VMXML.new_from_dumpxml(vm_name)
        devices = vmxml.devices
        graphics = devices.by_device_tag('graphics')[0]
        graphics.type_name = graphic
        if passwd is not None:
            graphics.passwd = passwd
        vmxml.devices = devices
        vmxml.sync()

    @staticmethod
    def add_graphic(vm_name, passwd=None, graphic="vnc",
                    add_channel=False):
        """
        Add spice ssl or vnc graphic with passwd

        :param vm_name: name of vm
        :param passwd: password for graphic
        :param graphic: graphic type, spice or vnc
        :param add_channel: add channel for spice
        """
        vmxml = vm_xml.VMXML.new_from_dumpxml(vm_name)
        grap = vmxml.get_device_class('graphics')(type_name=graphic)
        if passwd is not None:
            grap.passwd = passwd
        grap.autoport = "yes"
        if graphic == "spice" and add_channel:
            grap.add_channel(name='main', mode='secure')
            grap.add_channel(name='inputs', mode='secure')
        vmxml.devices = vmxml.devices.append(grap)
        vmxml.sync()

    @staticmethod
    def del_graphic(vm_name):
        """
        Del original graphic device

        :param vm_name: name of vm
        """
        vmxml = vm_xml.VMXML.new_from_dumpxml(vm_name)
        vmxml.xmltreefile.remove_by_xpath('/devices/graphics', remove_all=True)
        vmxml.xmltreefile.write()
        vmxml.sync()



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatsecret.html
"""

from virttest.libvirt_xml import base, accessors
from virttest.libvirt_xml import xcepts


class SecretXMLBase(base.LibvirtXMLBase):

    """
    Accessor methods for SecretXML class.

    Properties:
        secret_ephemeral:
            yes or no, operates on XML secret tag
        secret_private:
            yes or no, operates on XML secret tag
        description:
            string, operates on description tag
        uuid:
            string, operates on uuid tag
        auth_type:
            string, sercet authentication type, operates
            on auth tag
        auth_username:
            string, secret authentication username, operates
            on auth tag
        usage:
            string, operates on usage tag
        target:
            string, sub-tag of the usage tag, operates on
            target tag
        volume:
            the volume file path, sub-tag of the usage tag,
            operates on volume tag
    """

    __slots__ = ('secret_ephemeral', 'secret_private', 'description',
                 'auth_type', 'auth_username', 'uuid', 'usage', 'target',
                 'volume', 'usage_name')

    __uncompareable__ = base.LibvirtXMLBase.__uncompareable__

    __schema_name__ = "secret"

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLAttribute('secret_ephemeral', self, parent_xpath='/',
                               tag_name='secret', attribute='ephemeral')
        accessors.XMLAttribute('secret_private', self, parent_xpath='/',
                               tag_name='secret', attribute='private')
        accessors.XMLElementText('uuid', self, parent_xpath='/',
                                 tag_name='uuid')
        accessors.XMLAttribute('auth_type', self, parent_xpath='/',
                               tag_name='auth', attribute='type')
        accessors.XMLAttribute('auth_username', self, parent_xpath='/',
                               tag_name='auth', attribute='username')
        accessors.XMLElementText('description', self, parent_xpath='/',
                                 tag_name='description')
        accessors.XMLAttribute('usage', self, parent_xpath='/',
                               tag_name='usage', attribute='type')
        accessors.XMLElementText('usage_name', self, parent_xpath='/usage',
                                 tag_name='name')
        accessors.XMLElementText('target', self, parent_xpath='/usage',
                                 tag_name='target')
        accessors.XMLElementText('volume', self, parent_xpath='/usage',
                                 tag_name='volume')
        super(SecretXMLBase, self).__init__(virsh_instance=virsh_instance)


class SecretXML(SecretXMLBase):

    """
    Manipulators of a secret through it's XML definition.
    """

    __slots__ = []

    def __init__(self, ephemeral='yes', private='no',
                 virsh_instance=base.virsh):
        """
        Initialize new instance with empty XML
        """
        super(SecretXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = u"<secret ephemeral='%s' private='%s'><description>\
                     </description></secret>" % (ephemeral, private)

    @staticmethod
    def new_from_secret_dumpxml(uuid, virsh_instance=base.virsh):
        """
        Return new SecretXML instance from virsh secret-dumpxml command

        :param uuid: secret's uuid
        :param virsh_instance: virsh module or instance to use
        :return: New initialized SecretXML instance
        """
        secret_xml = SecretXML(virsh_instance=virsh_instance)
        result = virsh_instance.secret_dumpxml(uuid)
        secret_xml['xml'] = result.stdout_text.strip()

        return secret_xml

    @staticmethod
    def get_secret_details_by_uuid(uuid, virsh_instance=base.virsh):
        """
        Return secret XML by secret's uuid

        :param uuid: secret's uuid
        :return: secret XML dictionary
        """
        secret_xml = {}
        sec_xml = SecretXML.new_from_secret_dumpxml(uuid, virsh_instance)

        secret_xml['secret_ephemeral'] = sec_xml.secret_ephemeral
        secret_xml['secret_private'] = sec_xml.secret_private
        secret_xml['uuid'] = sec_xml.uuid
        secret_xml['description'] = sec_xml.description
        # secret XML may not has usage, target or volume tag
        try:
            secret_xml['usage'] = sec_xml.usage
        except xcepts.LibvirtXMLNotFoundError:
            secret_xml['usage'] = ""
        try:
            secret_xml['target'] = sec_xml.target
        except xcepts.LibvirtXMLNotFoundError:
            secret_xml['target'] = ""
        try:
            secret_xml['volume'] = sec_xml.volume
        except xcepts.LibvirtXMLNotFoundError:
            secret_xml['volume'] = ""
        return secret_xml



"""
filesystem device support class(es)

https://libvirt.org/formatdomain.html#elementsFilesystem
"""

from virttest.libvirt_xml.devices import base
from virttest.libvirt_xml import accessors


class Filesystem(base.TypedDeviceBase):

    __slots__ = ('accessmode', 'source', 'target', 'driver', 'binary', 'alias')

    def __init__(self, type_name='mount', virsh_instance=base.base.virsh):
        accessors.XMLAttribute('accessmode', self, parent_xpath='/',
                               tag_name='filesystem', attribute='accessmode')
        accessors.XMLElementDict('source', self, parent_xpath='/',
                                 tag_name='source')
        accessors.XMLElementDict('target', self, parent_xpath='/',
                                 tag_name='target')
        accessors.XMLElementDict('driver', self,
                                 parent_xpath='/',
                                 tag_name='driver')
        accessors.XMLElementNest('binary', self, parent_xpath='/',
                                 tag_name='binary', subclass=self.Binary,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementDict('alias', self,
                                 parent_xpath='/',
                                 tag_name='alias')
        super(Filesystem, self).__init__(device_tag='filesystem',
                                         type_name=type_name,
                                         virsh_instance=virsh_instance)

    class Binary(base.base.LibvirtXMLBase):

        """
        Filesystem binary subclass
        Typical xml looks like:
        <binary path='/usr/libexec/virtiofsd' xattr='on'>
            <cache mode='always'/>
            <lock posix='on' flock='on'/>
        </binary>
        """
        __slots__ = ('path', 'xattr', 'cache_mode', 'lock_posix', 'flock')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('path', self,
                                   parent_xpath='/',
                                   tag_name='binary',
                                   attribute='path')
            accessors.XMLAttribute('xattr', self,
                                   parent_xpath='/',
                                   tag_name='binary',
                                   attribute='xattr')
            accessors.XMLAttribute('cache_mode', self,
                                   parent_xpath='/',
                                   tag_name='cache',
                                   attribute='mode')
            accessors.XMLAttribute('lock_posix', self,
                                   parent_xpath='/',
                                   tag_name='lock',
                                   attribute='posix')
            accessors.XMLAttribute('flock', self,
                                   parent_xpath='/',
                                   tag_name='lock',
                                   attribute='flock')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<binary/>'



import time
import threading
import logging
import re
import signal

from avocado.utils import process
from avocado.utils import path as utils_path
from avocado.utils import distro
from avocado.core import exceptions

from virttest import remote
from virttest import virsh
from virttest import utils_misc
from virttest import test_setup
from virttest import utils_net
from virttest import utils_iptables
from virttest.utils_test import libvirt


# Migration Relative functions##############
class MigrationTest(object):

    """Class for migration tests"""

    def __init__(self):
        # To get result in thread, using member parameters
        # Result of virsh migrate command
        # True means command executed successfully
        self.RET_MIGRATION = True
        # A lock for threads
        self.RET_LOCK = threading.RLock()
        # The time spent when migrating vms
        # format: vm_name -> time(seconds)
        self.mig_time = {}
        # The CmdResult returned from virsh migrate command
        self.ret = None

    def post_migration_check(self, vms, params, uptime, uri=None):
        """
        Validating migration by performing checks in this method
        * uptime of the migrated vm > uptime of vm before migration
        * ping vm from target host
        * check vm state after migration

        :param vms: VM objects of migrating vms
        :param uptime: uptime dict of vms before migration
        :param uri: target virsh uri
        :return: updated dict of uptime
        """
        vm_state = params.get("virsh_migrated_state", "running")
        ping_count = int(params.get("ping_count", 10))
        for vm in vms:
            if not libvirt.check_vm_state(vm.name, vm_state, uri=uri):
                raise exceptions.TestFail("Migrated VMs failed to be in %s "
                                          "state at destination" % vm_state)
            logging.info("Guest state is '%s' at destination is as expected",
                         vm_state)
            if "offline" not in params.get("migrate_options"):
                vm_uptime = vm.uptime(connect_uri=uri)
                logging.info("uptime of migrated VM %s: %s", vm.name,
                             vm_uptime)
                if vm_uptime < uptime[vm.name]:
                    raise exceptions.TestFail("vm went for a reboot during "
                                              "migration")
                self.ping_vm(vm, params, uri=uri, ping_count=ping_count)
                # update vm uptime to check when migrating back
                uptime[vm.name] = vm_uptime
                vm.verify_dmesg(connect_uri=uri)
        return uptime

    def ping_vm(self, vm, params, uri=None, ping_count=10,
                ping_timeout=60):
        """
        Method used to ping the VM before and after migration

        :param vm: VM object
        :param params: Test dict params
        :param uri: connect uri
        :param ping_count: count of icmp packet
        :param ping_timeout: Timeout for the ping command
        """
        vm_ip = params.get("vm_ip_dict", {})
        server_session = None
        func = exceptions.TestError
        if uri and uri != "qemu:///system":
            func = exceptions.TestFail
            uri_backup = vm.connect_uri
            vm.connect_uri = uri
            server_session = test_setup.remote_session(params)
            # after migration VM would take some time to respond and to
            # avoid the race of framework querying IP address before VM
            # starts responding, provide timeout for 240 seconds to retry
            # and raise if VM fails to respond
            vm_ip[vm.name] = vm.get_address(session=server_session,
                                            timeout=240)
            logging.info("Check VM network connectivity after migrating")
        else:
            logging.info("Check VM network connectivity before migration")
            if not vm.is_alive():
                vm.start()
            vm.wait_for_login()
            vm_ip[vm.name] = vm.get_address()
            params["vm_ip_dict"] = vm_ip
        s_ping, o_ping = utils_net.ping(vm_ip[vm.name], count=ping_count,
                                        timeout=ping_timeout,
                                        output_func=logging.debug,
                                        session=server_session)
        logging.info(o_ping)
        if uri and uri != 'qemu:///system':
            vm.connect_uri = uri_backup
            if server_session:
                server_session.close()
        if s_ping != 0:
            if uri:
                if "offline" in params.get("migrate_options"):
                    logging.info("Offline Migration: %s will not responded to "
                                 "ping as expected", vm.name)
                    return
            func("%s did not respond after %d sec." % (vm.name, ping_timeout))

    def thread_func_migration(self, vm, desturi, options=None,
                              ignore_status=False, virsh_opt="",
                              extra_opts=""):
        """
        Thread for virsh migrate command.

        :param vm: A libvirt vm instance(local or remote).
        :param desturi: Remote host uri.
        :param options: The options for migration command.
        :param ignore_status: True, means no CmdError will be caught
                              for the failure.
                              False, means an CmdError will be caught
                              for the failure.
        """
        # Migrate the domain.
        is_error = False

        try:
            if options is None:
                options = "--live --timeout=60"
            stime = int(time.time())
            self.ret = vm.migrate(desturi, option=options,
                                  ignore_status=ignore_status,
                                  debug=True, virsh_opt=virsh_opt,
                                  extra=extra_opts)
            etime = int(time.time())
            self.mig_time[vm.name] = etime - stime
            if self.ret.exit_status != 0:
                logging.debug("Migration to %s returns failed exit status %d",
                              desturi, self.ret.exit_status)
                is_error = True
        except process.CmdError as detail:
            logging.error("Migration to %s failed:\n%s", desturi, detail)
            is_error = True
        finally:
            if is_error is True:
                self.RET_LOCK.acquire()
                self.RET_MIGRATION = False
                self.RET_LOCK.release()

    def migrate_pre_setup(self, desturi, params,
                          cleanup=False,
                          ports='49152:49216'):
        """
        # Setup before migration,
        # 1. To enable migration ports using iptables
        # 2. Turn off SMT for power8 machine in remote machine to migrate

        :param desturi: uri of destination machine to which VM gets migrated
        :param params: Test params dict
        :param cleanup: if True revert back to default setting, used to cleanup
        :param ports: ports used for allowing migration
        """
        use_firewall_cmd = distro.detect().name != "Ubuntu"
        iptables_func = utils_iptables.Iptables.setup_or_cleanup_iptables_rules
        try:
            utils_path.find_command("firewall-cmd")
        except utils_path.CmdNotFoundError:
            logging.debug("Using iptables for replacement")
            use_firewall_cmd = False

        if use_firewall_cmd:
            port_to_add = ports
            if ":" in ports:
                port_to_add = "%s-%s" % (ports.split(":")[0], ports.split(":")[1])
        else:
            rule = ["INPUT -p tcp -m tcp --dport %s -j ACCEPT" % ports]

        try:
            dest_ip = re.search(r'//.*/', desturi,
                                re.I).group(0).strip('/').strip()
            source_ip = params.get("migrate_source_host", "").strip()
            source_cn = params.get("migrate_source_host_cn", "").strip()
            # check whether migrate back to source machine or not
            if ((desturi == "qemu:///system") or (dest_ip == source_ip) or (dest_ip == source_cn)):
                if use_firewall_cmd:
                    firewall_cmd = utils_iptables.Firewall_cmd()
                    if cleanup:
                        firewall_cmd.remove_port(port_to_add, 'tcp', permanent=True)
                    else:
                        firewall_cmd.add_port(port_to_add, 'tcp', permanent=True)
                    # open migration ports in local machine using firewall_cmd
                else:
                    # open migration ports in local machine using iptables
                    iptables_func(rule, cleanup=cleanup)
                # SMT for Power8 machine is turned off for local machine during
                # test setup
            else:
                server_ip = params.get("server_ip", params.get("remote_ip"))
                server_user = params.get("server_user", params.get("remote_user"))
                server_pwd = params.get("server_pwd", params.get("remote_pwd"))
                server_session = remote.wait_for_login('ssh', server_ip, '22',
                                                       server_user, server_pwd,
                                                       r"[\#\$]\s*$")
                if use_firewall_cmd:
                    firewall_cmd = utils_iptables.Firewall_cmd(server_session)
                    # open migration ports in remote machine using firewall_cmd
                    if cleanup:
                        firewall_cmd.remove_port(port_to_add, 'tcp', permanent=True)
                    else:
                        firewall_cmd.add_port(port_to_add, 'tcp', permanent=True)
                else:
                    # open migration ports in remote machine using iptables
                    iptables_func(rule, params=params, cleanup=cleanup)
                cmd = "grep cpu /proc/cpuinfo | awk '{print $3}' | head -n 1"
                # Check if remote machine is Power8, if so check for smt state
                # and turn off if it is on.
                cmd_output = server_session.cmd_status_output(cmd)
                server_session.close()
                if (cmd_output[0] == 0):
                    cmd_output = cmd_output[1].strip().upper()
                    if "POWER8" in cmd_output:
                        test_setup.switch_smt(state="off", params=params)
                else:
                    raise exceptions.TestError("Failed to get cpuinfo of remote "
                                               "server", cmd_output[1])
        except AttributeError:
            # Negative scenarios will have invalid desturi for which test should
            # continue
            pass

    def do_migration(self, vms, srcuri, desturi, migration_type,
                     options=None, thread_timeout=60,
                     ignore_status=False, func=None, virsh_opt="",
                     extra_opts="", **args):
        """
        Migrate vms.

        :param vms: migrated vms.
        :param srcuri: local uri, used when migrate vm from remote to local
        :param descuri: remote uri, used when migrate vm from local to remote
        :param migration_type: do orderly for simultaneous migration
        :param options: migration options
        :param thread_timeout: time out seconds for the migration thread running
        :param ignore_status: determine if an exception is raised for errors
        :param func: the function executed during migration thread is running
        :param args: dictionary used by func,
                     'func_param' is mandatory if no real func_param, none is
                     requested.
                     'shell' is optional, where shell=True(bool) can be used
                     for process.run

        """
        for vm in vms:
            vm.connect_uri = args.get("virsh_uri", "qemu:///system")
        if migration_type == "orderly":
            for vm in vms:
                migration_thread = threading.Thread(target=self.thread_func_migration,
                                                    args=(vm, desturi, options,
                                                          ignore_status, virsh_opt,
                                                          extra_opts))
                migration_thread.start()
                eclipse_time = 0
                stime = int(time.time())
                if func:
                    # Execute command once the migration is started
                    migrate_start_state = args.get("migrate_start_state", "paused")

                    # Wait for migration to start
                    migrate_options = ""
                    if options:
                        migrate_options = str(options)
                    if extra_opts:
                        migrate_options += " %s" % extra_opts

                    migration_started = self.wait_for_migration_start(vm, state=migrate_start_state,
                                                                      uri=desturi,
                                                                      migrate_options=migrate_options.strip())

                    if migration_started:
                        logging.info("Migration started for %s", vm.name)
                        if func == process.run:
                            try:
                                func(args['func_params'], shell=args['shell'])
                            except KeyError:
                                func(args['func_params'])
                        elif func == virsh.migrate_postcopy:
                            time.sleep(3)  # To avoid of starting postcopy before starting migration
                            func(vm.name, uri=srcuri, debug=True)
                        else:
                            if 'func_params' in args:
                                func(args['func_params'])
                            else:
                                func()
                    else:
                        logging.error("Migration failed to start for %s",
                                      vm.name)
                eclipse_time = int(time.time()) - stime
                logging.debug("start_time:%d, eclipse_time:%d", stime, eclipse_time)
                if eclipse_time < thread_timeout:
                    migration_thread.join(thread_timeout - eclipse_time)
                if migration_thread.isAlive():
                    logging.error("Migrate %s timeout.", migration_thread)
                    self.RET_LOCK.acquire()
                    self.RET_MIGRATION = False
                    self.RET_LOCK.release()
        elif migration_type == "cross":
            # Migrate a vm to remote first,
            # then migrate another to remote with the first vm back
            vm_remote = vms.pop()
            self.thread_func_migration(vm_remote, desturi)
            for vm in vms:
                thread1 = threading.Thread(target=self.thread_func_migration,
                                           args=(vm_remote, srcuri, options))
                thread2 = threading.Thread(target=self.thread_func_migration,
                                           args=(vm, desturi, options))
                thread1.start()
                thread2.start()
                thread1.join(thread_timeout)
                thread2.join(thread_timeout)
                vm_remote = vm
                if thread1.isAlive() or thread1.isAlive():
                    logging.error("Cross migrate timeout.")
                    self.RET_LOCK.acquire()
                    self.RET_MIGRATION = False
                    self.RET_LOCK.release()
            # Add popped vm back to list
            vms.append(vm_remote)
        elif migration_type == "simultaneous":
            migration_threads = []
            for vm in vms:
                migration_threads.append(threading.Thread(
                                         target=self.thread_func_migration,
                                         args=(vm, desturi, options)))
            # let all migration going first
            for thread in migration_threads:
                thread.start()

            # listen threads until they end
            for thread in migration_threads:
                thread.join(thread_timeout)
                if thread.isAlive():
                    logging.error("Migrate %s timeout.", thread)
                    self.RET_LOCK.acquire()
                    self.RET_MIGRATION = False
                    self.RET_LOCK.release()
        if not self.RET_MIGRATION and not ignore_status:
            raise exceptions.TestFail()

    def cleanup_dest_vm(self, vm, srcuri, desturi):
        """
        Cleanup migrated vm on remote host.
        """
        vm.connect_uri = desturi
        if vm.exists():
            if vm.is_persistent():
                vm.undefine()
            if vm.is_alive():
                # If vm on remote host is unaccessible
                # graceful shutdown may cause confused
                vm.destroy(gracefully=False)
        # Set connect uri back to local uri
        vm.connect_uri = srcuri

    def wait_for_migration_start(self, vm, state='paused', uri=None,
                                 migrate_options='', timeout=60):
        """
        checks whether migration is started or not

        :param vm: VM object
        :param state: expected VM state in destination host
        :param uri: connect uri
        :param timeout: time in seconds to wait for migration to start
        :param migrate_options: virsh migrate options

        :return: True if migration is started False otherwise
        """
        def check_state():
            try:
                return libvirt.check_vm_state(dest_vm_name, state, uri=uri)
            except Exception:
                return False

        # Set dest_vm_name to be used in wait_for_migration_start() in case
        # --dname is specified in virsh options
        dest_vm_name = ""
        if migrate_options.count("--dname"):
            migrate_options_list = migrate_options.split()
            dest_vm_name = migrate_options_list[migrate_options_list.index("--dname") + 1]
        else:
            dest_vm_name = vm.name

        return utils_misc.wait_for(check_state, timeout)

    def check_parameters(self, params):
        """
        Make sure all of parameters are assigned a valid value

        :param params: the parameters to be checked

        :raise: test.cancel if invalid value exists
        """
        migrate_dest_host = params.get("migrate_dest_host")
        migrate_dest_pwd = params.get("migrate_dest_pwd")
        migrate_source_host = params.get("migrate_source_host")
        migrate_source_pwd = params.get("migrate_source_pwd")

        args_list = [migrate_dest_host,
                     migrate_dest_pwd, migrate_source_host,
                     migrate_source_pwd]

        for arg in args_list:
            if arg and arg.count("EXAMPLE"):
                raise exceptions.TestCancel("Please assign a value for %s!" % arg)

    def check_result(self, result, params):
        """
        Check if the migration result is as expected

        :param result: the output of migration
        :param params: the parameters dict
        :raise: test.fail if test is failed
        """
        status_error = "yes" == params.get("status_error", "no")
        err_msg = params.get("err_msg")
        if not result:
            raise exceptions.TestError("No migration result is returned.")

        logging.info("Migration out: %s", result.stdout_text.strip())
        logging.info("Migration error: %s", result.stderr_text.strip())

        if status_error:  # Migration should fail
            if err_msg:   # Special error messages are expected
                if not re.search(err_msg, result.stderr_text.strip()):
                    raise exceptions.TestFail("Can not find the expected "
                                              "patterns '%s' in output '%s'"
                                              % (err_msg,
                                                 result.stderr_text.strip()))
                else:
                    logging.debug("It is the expected error message")
            else:
                if int(result.exit_status) != 0:
                    logging.debug("Migration failure is expected result")
                else:
                    raise exceptions.TestFail("Migration success is unexpected result")
        else:
            if int(result.exit_status) != 0:
                raise exceptions.TestFail(result.stderr_text.strip())

    def do_cancel(self, sig=signal.SIGKILL):
        """
        Kill process during migration.

        :param sig: The signal to send
        :raise: test.error when kill fails
        """
        def _get_pid():
            cmd = "ps aux |grep 'virsh .*migrate' |grep -v grep |awk '{print $2}'"
            pid = process.run(cmd, shell=True).stdout_text
            return pid

        pid = utils_misc.wait_for(_get_pid, 30)
        if utils_misc.safe_kill(pid, sig):
            logging.info("Succeed to cancel migration: [%s].", pid.strip())
        else:
            raise exceptions.TestError("Fail to cancel migration: [%s]"
                                       % pid.strip())



import os
# pylint: disable=E0611
import distutils.ccompiler
import distutils.sysconfig

from virttest import data_dir

PYTHON_HEADERS = distutils.sysconfig.get_python_inc()
PYTHON_VERSION = distutils.sysconfig.get_python_version()
PYTHON_LIB = "python%s" % PYTHON_VERSION
if float(PYTHON_VERSION) >= 3:
    PYTHON_LIB += "m"

OUTPUT_DIR = os.path.dirname(__file__)

SOURCES = [os.path.join(OUTPUT_DIR, f) for f in ['passfd.c']]
SHARED_OBJECT = '_passfd.so'


def passfd_setup(output_dir=OUTPUT_DIR):
    '''
    Compiles the passfd python extension.

    :param output_dir: where the _passfd.so module will be saved
    :return: None
    '''
    if output_dir is None:
        output_dir = OUTPUT_DIR

    output_file = os.path.join(output_dir, SHARED_OBJECT)

    c = distutils.ccompiler.new_compiler()
    distutils.sysconfig.customize_compiler(c)
    objects = c.compile(SOURCES, include_dirs=[PYTHON_HEADERS],
                        output_dir=data_dir.get_tmp_dir(),
                        extra_postargs=['-fPIC'])
    c.link_shared_object(objects, output_file, libraries=[PYTHON_LIB])


def import_passfd():
    '''
    Imports and lazily sets up the passfd module

    :return: passfd module
    '''
    try:
        from virttest import passfd
    except ImportError:
        passfd_setup()
        from virttest import passfd

    return passfd


if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        passfd_setup(sys.argv[1])
    else:
        passfd_setup()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Auxiliary script used to allocate memory on guests.

:copyright: 2008-2009 Red Hat Inc.
:author: Jiri Zupka (jzupka@redhat.com)
"""

from __future__ import division
import os
import array
import sys
import random
import copy
import tempfile
import datetime
import math

from six.moves import input


PAGE_SIZE = 4096  # machine page size

TMPFS_OVERHEAD = 0.0022  # overhead on 1MB of write data


class MemFill(object):

    """
    Fills guest memory according to certain patterns.
    """

    def __init__(self, mem, static_value, random_key):
        """
        Constructor of MemFill class.

        :param mem: Amount of test memory in MB.
        :param random_key: Seed of random series used for fill up memory.
        :param static_value: Value used to fill all memory.
        """
        if (static_value < 0 or static_value > 255):
            print("FAIL: Initialization static value"
                  "can be only in range (0..255)")
            return

        self.tmpdp = tempfile.mkdtemp()
        tmpfs_size = mem + math.ceil(mem * TMPFS_OVERHEAD)
        ret_code = os.system("mount -o size=%dM tmpfs %s -t tmpfs" %
                             (tmpfs_size, self.tmpdp))
        if ret_code != 0:
            if os.getuid() != 0:
                print("FAIL: Unable to mount tmpfs "
                      "(likely cause: you are not root)")
            else:
                print("FAIL: Unable to mount tmpfs")
        else:
            self.f = tempfile.TemporaryFile(prefix='mem', dir=self.tmpdp)
            self.allocate_by = 'L'
            self.npages = ((mem * 1024 * 1024) // PAGE_SIZE)
            self.random_key = random_key
            self.static_value = static_value
            print("PASS: Initialization (tmpfs size: %dM)" % tmpfs_size)

    def __del__(self):
        if os.path.ismount(self.tmpdp):
            self.f.close()
            os.system("umount %s" % (self.tmpdp))

    def compare_page(self, original, inmem):
        """
        Compare pages of memory and print the differences found.

        :param original: Data that was expected to be in memory.
        :param inmem: Data in memory.
        """
        for ip in range(PAGE_SIZE // original.itemsize):
            if (not original[ip] == inmem[ip]):  # find which item is wrong
                originalp = array.array("B")
                inmemp = array.array("B")
                originalp.fromstring(original[ip:ip + 1].tostring())
                inmemp.fromstring(inmem[ip:ip + 1].tostring())
                for ib in range(len(originalp)):  # find wrong byte in item
                    if not (originalp[ib] == inmemp[ib]):
                        position = (self.f.tell() - PAGE_SIZE + ip *
                                    original.itemsize + ib)
                        print("Mem error on position %d wanted 0x%Lx and is "
                              "0x%Lx" % (position, originalp[ib], inmemp[ib]))

    def value_page(self, value):
        """
        Create page filled by value.

        :param value: String we want to fill the page with.
        :return: return array of bytes size PAGE_SIZE.
        """
        a = array.array("B")
        for _ in range((PAGE_SIZE // a.itemsize)):
            try:
                a.append(value)
            except Exception:
                print("FAIL: Value can be only in range (0..255)")
        return a

    def random_page(self, seed):
        """
        Create page filled by static random series.

        :param seed: Seed of random series.
        :return: Static random array series.
        """
        random.seed(seed)
        a = array.array(self.allocate_by)
        for _ in range(PAGE_SIZE // a.itemsize):
            a.append(random.randrange(0, sys.maxsize))
        return a

    def value_fill(self, value=None):
        """
        Fill memory page by page, with value generated with value_page.

        :param value: Parameter to be passed to value_page. None to just use
                what's on the attribute static_value.
        """
        self.f.seek(0)
        if value is None:
            value = self.static_value
        page = self.value_page(value)
        for _ in range(self.npages):
            page.tofile(self.f)
        print("PASS: Mem value fill")

    def value_check(self, value=None):
        """
        Check memory to see if data is correct.

        :param value: Parameter to be passed to value_page. None to just use
                what's on the attribute static_value.
        :return: if data in memory is correct return PASS
                else print some wrong data and return FAIL
        """
        self.f.seek(0)
        e = 2
        failure = False
        if value is None:
            value = self.static_value
        page = self.value_page(value)
        for _ in range(self.npages):
            pf = array.array("B")
            pf.fromfile(self.f, PAGE_SIZE // pf.itemsize)
            if not (page == pf):
                failure = True
                self.compare_page(page, pf)
                e = e - 1
                if e == 0:
                    break
        if failure:
            print("FAIL: value verification")
        else:
            print("PASS: value verification")

    def static_random_fill(self, n_bytes_on_end=PAGE_SIZE):
        """
        Fill memory by page with static random series with added special value
        on random place in pages.

        :param n_bytes_on_end: how many bytes on the end of page can be changed.
        :return: PASS.
        """
        self.f.seek(0)
        page = self.random_page(self.random_key)
        random.seed(self.random_key)
        p = copy.copy(page)

        t_start = datetime.datetime.now()
        for pages in range(self.npages):
            rand = random.randint(((PAGE_SIZE // page.itemsize) - 1) -
                                  (n_bytes_on_end // page.itemsize),
                                  (PAGE_SIZE // page.itemsize) - 1)
            p[rand] = pages
            p.tofile(self.f)
            p[rand] = page[rand]

        t_end = datetime.datetime.now()
        delta = t_end - t_start
        milisec = delta.microseconds // 1e3 + delta.seconds * 1e3
        print("PASS: filling duration = %Ld ms" % milisec)

    def static_random_verify(self, n_bytes_on_end=PAGE_SIZE):
        """
        Check memory to see if it contains correct contents.

        :return: if data in memory is correct return PASS
                else print some wrong data and return FAIL.
        """
        self.f.seek(0)
        e = 2
        page = self.random_page(self.random_key)
        random.seed(self.random_key)
        p = copy.copy(page)
        failure = False
        for pages in range(self.npages):
            rand = random.randint(((PAGE_SIZE // page.itemsize) - 1) -
                                  (n_bytes_on_end // page.itemsize),
                                  (PAGE_SIZE // page.itemsize) - 1)
            p[rand] = pages
            pf = array.array(self.allocate_by)
            pf.fromfile(self.f, PAGE_SIZE // pf.itemsize)
            if not (p == pf):
                failure = True
                self.compare_page(p, pf)
                e = e - 1
                if e == 0:
                    break
            p[rand] = page[rand]
        if failure:
            print("FAIL: Random series verification")
        else:
            print("PASS: Random series verification")


def die():
    """
    Quit allocator.
    """
    sys.exit(0)


def main():
    """
    Main (infinite) loop of allocator.
    """
    print("PASS: Start")
    end = False
    while not end:
        sr = input()
        exec(sr)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Helpers for cgroup testing.

:copyright: 2011 Red Hat Inc.
:author: Lukas Doktor <ldoktor@redhat.com>
"""
import logging
import os
import shutil
import subprocess
import time
import re
import random
import six
from tempfile import mkdtemp

from avocado.core import exceptions
from avocado.utils import software_manager
from avocado.utils import process

from . import service


class Cgroup(object):

    """
    Cgroup handling class.
    """

    def __init__(self, module, _client):
        """
        Constructor
        :param module: Name of the cgroup module
        :param _client: Test script pwd + name
        """
        self.module = module
        self._client = _client
        self.root = None
        self.cgroups = []

    def __del__(self):
        """
        Destructor
        """
        self.cgroups.sort(reverse=True)
        for pwd in self.cgroups[:]:
            for task in self.get_property("tasks", pwd):
                if task:
                    self.set_root_cgroup(int(task))
            self.rm_cgroup(pwd)

    def initialize(self, modules):
        """
        Initializes object for use.

        :param modules: Array of all available cgroup modules.
        """
        self.root = modules.get_pwd(self.module)
        if not self.root:
            raise exceptions.TestError("cg.initialize(): Module %s not found"
                                       % self.module)

    def __get_cgroup_pwd(self, cgroup):
        """
        Get cgroup's full path

        :param cgroup: cgroup name
        :return: cgroup's full path
        """
        if not isinstance(cgroup, six.string_types):
            raise exceptions.TestError("cgroup type isn't string!")
        return os.path.join(self.root, cgroup) + '/'

    def get_cgroup_name(self, pwd=None):
        """
        Get cgroup's name

        :param pwd: cgroup name
        :return: cgroup's name
        """
        if pwd is None:
            # root cgroup
            return None
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        # self.root is "/cgroup/blkio," not "/cgroup/blkio/"
        # cgroup is "/cgroup/blkio/test" or "/cgroup/blkio/test/test"
        # expected cgroup name is test or test/test
        if pwd.startswith(self.root + '/'):
            return pwd[len(self.root) + 1: -1]
        return None

    def get_cgroup_index(self, cgroup):
        """
        Get cgroup's index in cgroups

        :param cgroup: cgroup name
        :return: index of cgroup
        """
        try:
            if self.__get_cgroup_pwd(cgroup) not in self.cgroups:
                raise exceptions.TestFail("%s not exists!" % cgroup)
            cgroup_pwd = self.__get_cgroup_pwd(cgroup)
            return self.cgroups.index(cgroup_pwd)
        except process.CmdError:
            raise exceptions.TestFail("Find index failed!")

    def mk_cgroup_cgcreate(self, pwd=None, cgroup=None):
        """
        Make a cgroup by executing the cgcreate command

        :params: cgroup: name of the cgroup to be created
        :return: last cgroup index
        """
        try:
            parent_cgroup = self.get_cgroup_name(pwd)
            if cgroup is None:
                range = "abcdefghijklmnopqrstuvwxyz0123456789"
                sub_cgroup = "cgroup-" + "".join(random.sample(range +
                                                               range.upper(), 6))
            else:
                sub_cgroup = cgroup
            if parent_cgroup is None:
                cgroup = sub_cgroup
            else:
                # Parent cgroup:test. Created cgroup:test1.
                # Whole cgroup name is "test/test1"
                cgroup = os.path.join(parent_cgroup, sub_cgroup)
            if self.__get_cgroup_pwd(cgroup) in self.cgroups:
                raise exceptions.TestFail("%s exists!" % cgroup)
            cgcreate_cmd = "cgcreate -g %s:%s" % (self.module, cgroup)
            process.run(cgcreate_cmd, ignore_status=False)
            pwd = self.__get_cgroup_pwd(cgroup)
            self.cgroups.append(pwd)
            return len(self.cgroups) - 1
        except process.CmdError:
            raise exceptions.TestFail("Make cgroup by cgcreate failed!")

    def mk_cgroup(self, pwd=None, cgroup=None):
        """
        Creates new temporary cgroup
        :param pwd: where to create this cgroup (default: self.root)
        :param cgroup: desired cgroup name
        :return: last cgroup index
        """
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            if cgroup and self.__get_cgroup_pwd(cgroup) in self.cgroups:
                raise exceptions.TestFail("%s exists!" % cgroup)
            if not cgroup:
                pwd = mkdtemp(prefix='cgroup-', dir=pwd) + '/'
            else:
                pwd = os.path.join(pwd, cgroup) + '/'
                if not os.path.exists(pwd):
                    os.mkdir(pwd)
        except Exception as inst:
            raise exceptions.TestError("cg.mk_cgroup(): %s" % inst)
        self.cgroups.append(pwd)
        return len(self.cgroups) - 1

    def cgexec(self, cgroup, cmd, args=""):
        """
        Execute command in desired cgroup

        :param cgroup: Desired cgroup
        :param cmd: Executed command
        :param args: Executed command's parameters
        """
        try:
            cgexec_cmd = ("cgexec -g %s:%s %s %s" %
                          (self.module, cgroup, cmd, args))
            status, output = process.getstatusoutput(cgexec_cmd)
            return status, output
        except process.CmdError as detail:
            raise exceptions.TestFail("Execute %s in cgroup failed!\n%s" %
                                      (cmd, detail))

    def rm_cgroup(self, pwd):
        """
        Removes cgroup.

        :param pwd: cgroup directory.
        """
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            os.rmdir(pwd)
            self.cgroups.remove(pwd)
        except ValueError:
            logging.warn("cg.rm_cgroup(): Removed cgroup which wasn't created"
                         "using this Cgroup")
        except Exception as inst:
            raise exceptions.TestError("cg.rm_cgroup(): %s" % inst)

    def get_all_cgroups(self):
        """
        Get all sub cgroups in this controller
        """
        lscgroup_cmd = "lscgroup %s:/" % self.module
        result = process.run(lscgroup_cmd, ignore_status=True)
        if result.exit_status:
            raise exceptions.TestFail(result.stderr_text.strip())
        cgroup_list = result.stdout_text.strip().splitlines()
        # Remove root cgroup
        cgroup_list = cgroup_list[1:]
        self.root = get_cgroup_mountpoint(self.module)
        sub_cgroup_list = []
        for item in cgroup_list:
            sub_cg = item.split(":/")[-1]
            sub_cg_path = os.path.join(self.root, sub_cg) + '/'
            sub_cgroup_list.append(sub_cg_path)
        self.cgroups = sub_cgroup_list
        return self.cgroups

    def cgdelete_all_cgroups(self):
        """
        Delete all cgroups in the module
        """
        try:
            for cgroup_pwd in self.cgroups:
                # Ignore sub cgroup
                cgroup = self.get_cgroup_name(cgroup_pwd)
                if cgroup.count("/") > 0:
                    continue
                self.cgdelete_cgroup(cgroup, True)
        except process.CmdError:
            raise exceptions.TestFail("cgdelete all cgroups in %s failed!"
                                      % self.module)

    def cgdelete_cgroup(self, cgroup, recursive=False):
        """
        Delete desired cgroup.

        :params cgroup: desired cgroup
        :params force: If true, sub cgroup can be deleted with parent cgroup
        """
        try:
            cgroup_pwd = self.__get_cgroup_pwd(cgroup)
            if cgroup_pwd not in self.cgroups:
                raise exceptions.TestError("%s doesn't exist!" % cgroup)
            cmd = "cgdelete %s:%s" % (self.module, cgroup)
            if recursive:
                cmd += " -r"
            process.run(cmd, ignore_status=False)
            self.cgroups.remove(cgroup_pwd)
        except process.CmdError as detail:
            raise exceptions.TestFail("cgdelete %s failed!\n%s" %
                                      (cgroup, detail))

    def cgclassify_cgroup(self, pid, cgroup):
        """
        Classify pid into cgroup

        :param pid: pid of the process
        :param cgroup: cgroup name
        """
        try:
            cgroup_pwd = self.__get_cgroup_pwd(cgroup)
            if cgroup_pwd not in self.cgroups:
                raise exceptions.TestError("%s doesn't exist!" % cgroup)
            cgclassify_cmd = ("cgclassify -g %s:%s %d" %
                              (self.module, cgroup, pid))
            process.run(cgclassify_cmd, ignore_status=False)
        except process.CmdError as detail:
            raise exceptions.TestFail("Classify process to tasks file "
                                      "failed!: %s" % detail)

    def get_pids(self, pwd=None):
        """
        Get all pids in cgroup

        :params: pwd: cgroup directory
        :return: all pids(list)
        """
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            return [_.strip() for _ in open(os.path.join(pwd, 'tasks'), 'r')]
        except Exception as inst:
            raise exceptions.TestError("cg.get_pids(): %s" % inst)

    def test(self, cmd):
        """
        Executes cgroup_client.py with cmd parameter.

        :param cmd: command to be executed
        :return: subprocess.Popen() process
        """
        logging.debug("cg.test(): executing parallel process '%s'", cmd)
        cmd = self._client + ' ' + cmd
        process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, close_fds=True,
                                   universal_newlines=True)
        return process

    def is_cgroup(self, pid, pwd):
        """
        Checks if the 'pid' process is in 'pwd' cgroup
        :param pid: pid of the process
        :param pwd: cgroup directory
        :return: 0 when is 'pwd' member
        """
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        if open(os.path.join(pwd, 'tasks')).readlines().count("%d\n" % pid) > 0:
            return 0
        else:
            return -1

    def is_root_cgroup(self, pid):
        """
        Checks if the 'pid' process is in root cgroup (WO cgroup)
        :param pid: pid of the process
        :return: 0 when is 'root' member
        """
        return self.is_cgroup(pid, self.root)

    def set_cgroup(self, pid, pwd=None):
        """
        Sets cgroup membership
        :param pid: pid of the process
        :param pwd: cgroup directory
        """
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            with open(os.path.join(pwd, 'tasks'), 'w') as tasks:
                tasks.write(str(pid))
        except Exception as inst:
            raise exceptions.TestError("cg.set_cgroup(): %s" % inst)
        if self.is_cgroup(pid, pwd):
            raise exceptions.TestError("cg.set_cgroup(): Setting %d pid into %s "
                                       "cgroup failed" % (pid, pwd))

    def set_root_cgroup(self, pid):
        """
        Resets the cgroup membership (sets to root)
        :param pid: pid of the process
        :return: 0 when PASSED
        """
        return self.set_cgroup(pid, self.root)

    def get_property(self, prop, pwd=None):
        """
        Gets the property value
        :param prop: property name (file)
        :param pwd: cgroup directory
        :return: [] values or None when FAILED
        """
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            # Remove tailing '\n' from each line
            file_link = os.path.join(pwd, prop)
            ret = [_[:-1].replace("\t", " ") for _ in open(file_link, 'r')]
            if ret:
                return ret
            else:
                return [""]
        except Exception as inst:
            raise exceptions.TestError("cg.get_property(): %s" % inst)

    def set_property_h(self, prop, value, pwd=None, check=True, checkprop=None):
        """
        Sets the one-line property value concerning the K,M,G postfix
        :param prop: property name (file)
        :param value: desired value
        :param pwd: cgroup directory
        :param check: check the value after setup / override checking value
        :param checkprop: override prop when checking the value
        """
        _value = value
        try:
            value = str(value)
            human = {'B': 1,
                     'K': 1024,
                     'M': 1048576,
                     'G': 1073741824,
                     'T': 1099511627776
                     }
            if value[-1] in human:
                value = int(value[:-1]) * human[value[-1]]
        except Exception:
            logging.warn("cg.set_prop() fallback into cg.set_property.")
            value = _value
        self.set_property(prop, value, pwd, check, checkprop)

    def set_property(self, prop, value, pwd=None, check=True, checkprop=None):
        """
        Sets the property value
        :param prop: property name (file)
        :param value: desired value
        :param pwd: cgroup directory
        :param check: check the value after setup / override checking value
        :param checkprop: override prop when checking the value
        """
        value = str(value)
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            open(os.path.join(pwd, prop), 'w').write(value)
        except Exception as inst:
            raise exceptions.TestError("cg.set_property(): %s" % inst)

        if check is not False:
            if check is True:
                check = value
            if checkprop is None:
                checkprop = prop
            _values = self.get_property(checkprop, pwd)
            # Sanitize non printable characters before check
            check = " ".join(check.split())
            if check not in _values:
                raise exceptions.TestError("cg.set_property(): Setting failed: "
                                           "desired = %s, real values = %s"
                                           % (repr(check), repr(_values)))

    def cgset_property(self, prop, value, pwd=None, check=True, checkprop=None):
        """
        Sets the property value by cgset command

        :param prop: property name (file)
        :param value: desired value
        :param pwd: cgroup directory
        :param check: check the value after setup / override checking value
        :param checkprop: override prop when checking the value
        """
        if pwd is None:
            pwd = self.root
        if isinstance(pwd, int):
            pwd = self.cgroups[pwd]
        try:
            cgroup = self.get_cgroup_name(pwd)
            cgset_cmd = "cgset -r %s='%s' %s" % (prop, value, cgroup)
            process.run(cgset_cmd, ignore_status=False)
        except process.CmdError as detail:
            raise exceptions.TestFail(
                "Modify %s failed!:\n%s" % (prop, detail))

        if check is not False:
            if check is True:
                check = value
            if checkprop is None:
                checkprop = prop
            _values = self.get_property(checkprop,
                                        self.get_cgroup_index(cgroup))
            # Sanitize non printable characters before check
            check = " ".join(check.split())
            if check not in _values:
                raise exceptions.TestError("cg.set_property(): Setting failed: "
                                           "desired = %s, real values = %s"
                                           % (repr(check), repr(_values)))

    def smoke_test(self):
        """
        Smoke test
        Module independent basic tests
        """
        pwd = self.mk_cgroup()

        ps = self.test("smoke")
        if ps is None:
            raise exceptions.TestError(
                "cg.smoke_test: Couldn't create process")

        if (ps.poll() is not None):
            raise exceptions.TestError(
                "cg.smoke_test: Process died unexpectidly")

        # New process should be a root member
        if self.is_root_cgroup(ps.pid):
            raise exceptions.TestError(
                "cg.smoke_test: Process is not a root member")

        # Change the cgroup
        self.set_cgroup(ps.pid, pwd)

        # Try to remove used cgroup
        try:
            self.rm_cgroup(pwd)
        except exceptions.TestError:
            pass
        else:
            raise exceptions.TestError("cg.smoke_test: Unexpected successful"
                                       " deletion of the used cgroup")

        # Return the process into the root cgroup
        self.set_root_cgroup(ps.pid)

        # It should be safe to remove the cgroup now
        self.rm_cgroup(pwd)

        # Finish the process
        ps.stdin.write('\n')
        time.sleep(2)
        if (ps.poll() is None):
            raise exceptions.TestError(
                "cg.smoke_test: Process is not finished")


class CgroupModules(object):

    """
    Handles the list of different cgroup filesystems.
    """

    def __init__(self, mountdir=None):
        self.modules = []
        self.modules.append([])
        self.modules.append([])
        self.modules.append([])
        if mountdir is None:
            self.mountdir = mkdtemp(prefix='cgroup-') + '/'
            self.rm_mountdir = True
        else:
            self.mountdir = mountdir
            self.rm_mountdir = False

    def __del__(self):
        """
        Unmount all cgroups and remove the mountdir
        """
        for i in range(len(self.modules[0])):
            if self.modules[2][i]:
                try:
                    process.system('umount %s -l' % self.modules[1][i])
                except Exception as failure_detail:
                    logging.warn("CGM: Couldn't unmount %s directory: %s",
                                 self.modules[1][i], failure_detail)
        try:
            if self.rm_mountdir:
                # If delete /cgroup/, this action will break cgroup service.
                shutil.rmtree(self.mountdir)
        except Exception:
            logging.warn(
                "CGM: Couldn't remove the %s directory", self.mountdir)

    def init(self, _modules):
        """
        Checks the mounted modules and if necessary mounts them into tmp
        mountdir.

        :param _modules: Desired modules.'memory','cpu,cpuset'...
        :return: Number of initialized modules.
        """
        logging.debug("Desired cgroup modules: %s", _modules)
        mounts = []
        with open('/proc/mounts', 'r') as proc_mounts:
            line = proc_mounts.readline().split()
            while line:
                if line[2] == 'cgroup':
                    mounts.append(line)
                line = proc_mounts.readline().split()

        for module in _modules:
            # Is it already mounted?
            i = False
            _module = set(module.split(','))
            for mount in mounts:
                # 'memory' or 'memory,cpuset'
                if _module.issubset(mount[3].split(',')):
                    self.modules[0].append(module)
                    self.modules[1].append(mount[1] + '/')
                    self.modules[2].append(False)
                    i = True
                    break
            if not i:
                # Not yet mounted
                module_path = os.path.join(self.mountdir, module)
                if not os.path.exists(module_path):
                    os.mkdir(module_path)
                cmd = ('mount -t cgroup -o %s %s %s' %
                       (module, module, module_path))
                try:
                    process.run(cmd)
                    self.modules[0].append(module)
                    self.modules[1].append(module_path)
                    self.modules[2].append(True)
                except process.CmdError:
                    logging.info("Cgroup module '%s' not available", module)

        logging.debug("Initialized cgroup modules: %s", self.modules[0])
        return len(self.modules[0])

    def get_pwd(self, module):
        """
        Returns the mount directory of 'module'
        :param module: desired module (memory, ...)
        :return: mount directory of 'module' or None
        """
        try:
            i = self.modules[0].index(module)
        except Exception as inst:
            logging.error("module %s not found: %s", module, inst)
            return None
        return self.modules[1][i]


def get_load_per_cpu(_stats=None):
    """
    Gather load per cpu from /proc/stat
    :param _stats: previous values
    :return: list of diff/absolute values of different CPU times for every cpu:
             [[user, nice, system, ...],[user, nice, system, ...], ...]
    """
    stats = []
    with open('/proc/stat', 'r') as f:
        f_stat = f.readlines()

    if _stats:
        for line in range(len(_stats)):
            l_stat = []
            for i in range(len(_stats[line])):
                l_stat.append(int(f_stat[line].split()[i+1]) - _stats[line][i])
            stats.append(l_stat)
    else:
        for line in f_stat:
            if line.startswith('cpu'):
                stats.append([int(v) for v in line.split()[1:]])
    return stats


def get_cgroup_mountpoint(controller, mount_file="/proc/mounts"):
    """
    Get desired controller's mountpoint

    :param controller: Desired controller
    :param mount_file: Name of file contains mounting information, in most
                       cases this are not need to be set.
    :return: controller's mountpoint
    :raise: TestError when contoller doesn't exist in mount table
    """
    f_cgcon = open(mount_file, "rU")
    cgconf_txt = f_cgcon.read()
    f_cgcon.close()
    mntpt = re.findall(
        r"\s(\S*cgroup/\S*%s(?=[,\ ])\S*)" % controller, cgconf_txt)
    if len(mntpt) == 0:
        # Controller is not supported if not found in mount table.
        raise exceptions.TestError(
            "Doesn't support controller <%s>" % controller)
    return mntpt[0]


def get_all_controllers():
    """
    Get all controllers used in system

    :return: all used controllers(controller_list)
    """
    try:
        result = process.run("lssubsys", ignore_status=False)
        controllers_str = result.stdout_text.strip()
        controller_list = []
        for controller in controllers_str.splitlines():
            controller_sub_list = controller.split(",")
            controller_list += controller_sub_list
    except process.CmdError:
        controller_list = ['cpuacct', 'cpu', 'memory', 'cpuset',
                           'devices', 'freezer', 'blkio', 'netcls']
    return controller_list


def resolve_task_cgroup_path(pid, controller):
    """
    Resolving cgroup mount path of a particular task

    :params: pid : process id of a task for which the cgroup path required
    :params: controller: takes one of the controller names in controller list

    :return: resolved path for cgroup controllers of a given pid
    """
    root_path = get_cgroup_mountpoint(controller)

    proc_cgroup = "/proc/%d/cgroup" % pid
    if not os.path.isfile(proc_cgroup):
        raise NameError('File %s does not exist\n Check whether cgroup \
                                    installed in the system' % proc_cgroup)

    try:
        proc_file = open(proc_cgroup, 'r')
        proc_cgroup_txt = proc_file.read()
    finally:
        proc_file.close()

    mount_path = re.findall(
        r":\S*%s(?=[,:])\S*:(\S*)\n" % controller, proc_cgroup_txt)
    return os.path.join(root_path, mount_path[0].strip("/"))


class CgconfigService(object):

    """
    Cgconfig service class.
    """

    def __init__(self):
        # libcgroup lack libcgroup-tools dependency will introduces
        # following error,
        #     Failed to issue method call:
        #     Unit cgconfig.service failed to load:
        #     No such file or directory
        #
        # Please refer to
        # https://bugzilla.redhat.com/show_bug.cgi?format=multiple&id=882887
        manager = software_manager.SoftwareManager()
        if not manager.install('libcgroup-tools'):
            exceptions.TestError("Failed to install libcgroup-tools on host")
        self._service_manager = service.Factory.create_service("cgconfig")

    def _service_cgconfig_control(self, action):
        """
        Cgconfig control by action.

        If cmd executes successfully, return True, otherwise return False.
        If the action is status, return True when it's running, otherwise return
        False.

        :param action: cgconfig service action
        """
        if not hasattr(self._service_manager, action):
            raise exceptions.TestError("Unknown action: %s" % action)
        return getattr(self._service_manager, action)()

    def cgconfig_start(self):
        """
        Sart cgconfig service
        """
        return self._service_cgconfig_control("start")

    def cgconfig_stop(self):
        """
        Sop cgconfig service
        """
        return self._service_cgconfig_control("stop")

    def cgconfig_restart(self):
        """
        Restart cgconfig service
        """
        return self._service_cgconfig_control("restart")

    def cgconfig_condrestart(self):
        """
        Condrestart cgconfig service
        """
        return self._service_cgconfig_control("condrestart")

    def cgconfig_is_running(self):
        """
        Check cgconfig service status
        """
        return self._service_cgconfig_control("status")


def all_cgroup_delete():
    """
    Clear all cgroups in system
    """
    try:
        process.run("cgclear", ignore_status=False)
    except process.CmdError as detail:
        raise exceptions.TestFail("Clear all cgroup failed!:\n%s" % detail)



"""
Installer classes are responsible for building and installing virtualization
specific software components. This is the main entry point for tests that
wish to install virtualization software components.

The most common use case is to simply call make_installer() inside your tests.
"""

from avocado.core import exceptions

from . import base_installer
from . import qemu_installer
from . import libvirt_installer

__all__ = ['InstallerRegistry', 'INSTALLER_REGISTRY', 'make_installer',
           'run_installers']


class InstallerRegistry(dict):

    """
    Holds information on known installer classes

    This class is used to create a single instance, named INSTALLER_REGISTRY,
    that will hold all information on known installer types.

    For registering a new installer class, use the register() method. If the
    virt type is not set explicitly, it will be set to 'base'. Example:

    >>> INSTALLER_REGISTRY.register('yum', base_installer.YumInstaller)

    If you want to register a virt specific installer class, set the virt
    (third) param:

    >>> INSTALLER_REGISTRY.register('yum', qemu_installer.YumInstaller, 'qemu')

    For getting a installer class, use the get_installer() method. This method
    has a fallback option 'get_default_virt' that will return a generic virt
    installer if set to true.
    """

    DEFAULT_VIRT_NAME = 'base'

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)
        self[self.DEFAULT_VIRT_NAME] = {}

    def register(self, mode, klass, virt=None):
        """
        Register a class as responsible for installing virt software components

        If virt is not set, it will assume a default of 'base'.
        """
        if virt is None:
            virt = self.DEFAULT_VIRT_NAME
        elif virt not in self:
            self[virt] = {}

        self[virt][mode] = klass

    def get_installer(self, mode, virt=None, get_default_virt=False):
        """
        Gets a installer class that should be able to install the virt software

        Always try to use classes that are specific to the virtualization
        technology that is being tested. If you have confidence that the
        installation is rather trivial and does not require custom steps, you
        may be able to get away with a base class (by setting get_default_virt
        to True).
        """
        if virt is None:
            virt = self.DEFAULT_VIRT_NAME
        if virt not in self:
            # return a base installer so the test could and give it a try?
            if get_default_virt:
                return self[self.DEFAULT_VIRT_NAME].get(mode)
        else:
            return self[virt].get(mode)

    def get_modes(self, virt=None):
        """
        Returns a list of all registered installer modes
        """
        if virt is None:
            virt = self.DEFAULT_VIRT_NAME

        if virt not in self:
            return []

        return list(self[virt].keys())


#
# InstallerRegistry unique instance
#
INSTALLER_REGISTRY = InstallerRegistry()


#
# Register base installers
#
INSTALLER_REGISTRY.register('yum',
                            base_installer.YumInstaller)
INSTALLER_REGISTRY.register('koji',
                            base_installer.KojiInstaller)
INSTALLER_REGISTRY.register('git_repo',
                            base_installer.GitRepoInstaller)
INSTALLER_REGISTRY.register('local_src',
                            base_installer.LocalSourceDirInstaller)
INSTALLER_REGISTRY.register('local_tar',
                            base_installer.LocalSourceTarInstaller)
INSTALLER_REGISTRY.register('remote_tar',
                            base_installer.RemoteSourceTarInstaller)

#
# Register KVM specific installers
#
INSTALLER_REGISTRY.register('yum',
                            base_installer.YumInstaller,
                            'qemu')
INSTALLER_REGISTRY.register('koji',
                            base_installer.KojiInstaller,
                            'qemu')
INSTALLER_REGISTRY.register('git_repo',
                            qemu_installer.GitRepoInstaller,
                            'qemu')
INSTALLER_REGISTRY.register('local_src',
                            qemu_installer.LocalSourceDirInstaller,
                            'qemu')
INSTALLER_REGISTRY.register('local_tar',
                            qemu_installer.LocalSourceTarInstaller,
                            'qemu')
INSTALLER_REGISTRY.register('remote_tar',
                            qemu_installer.RemoteSourceTarInstaller,
                            'qemu')

INSTALLER_REGISTRY.register('git_repo',
                            libvirt_installer.GitRepoInstaller,
                            'libvirt')


def installer_name_split(fullname, virt=None):
    """
    Split a full installer name into mode and short name

    Examples:
       git_repo_foo -> (git_repo, foo)
       local_src_foo -> (local_src, foo)
    """
    for mode in INSTALLER_REGISTRY.get_modes(virt):
        if fullname.startswith('%s_' % mode):
            _, _name = fullname.split(mode)
            name = _name[1:]
            return (mode, name)

    return (None, None)


def make_installer(fullname, params, test=None):
    """
    Installer factory: returns a new installer for the chosen mode and vm type

    This is the main entry point for acquiring an installer. Tests, such as
    the build test, should use this function.

    Param priority evaluation order is 'install_mode', then 'mode'. For virt
    type, 'vm_type' is consulted.

    :param fullname: the full name of instance, eg: git_repo_foo
    :param params: dictionary with parameters generated from cartersian config
    :param test: the test instance
    """
    virt = params.get("vm_type", None)

    mode, name = installer_name_split(fullname, virt)
    if mode is None or name is None:

        error_msg = ('Invalid installer mode or name for "%s". Probably an '
                     'installer has not been registered' % fullname)
        if virt is not None:
            error_msg += ' specifically for virt type "%s"' % virt

        raise exceptions.TestError(error_msg)

    klass = INSTALLER_REGISTRY.get_installer(mode, virt)
    if klass is None:
        raise exceptions.TestError(
            'Installer mode %s is not registered' % mode)
    else:
        return klass(mode, name, test, params)


def run_installers(params, test=None):
    """
    Runs the installation routines for all installers, one at a time

    This is usually the main entry point for tests
    """
    for name in params.get("installers", "").split():
        installer = make_installer(name, params, test)
        installer.install()
